from .detect_face import detect_face, create_mtcnn
import numpy as np

# modify the default parameters of np.load
np_load_old = np.load
np.load = lambda *a, **k: np_load_old(*a, allow_pickle=True, **k)


def to_rgb(img):
    w, h = img.shape
    ret = np.empty((w, h, 3), dtype=np.uint8)
    ret[:, :, 0] = ret[:, :, 1] = ret[:, :, 2] = img
    return ret


def aligner(sess):
    pnet, rnet, onet = create_mtcnn(sess, None)
    return [pnet, rnet, onet]


def align(orig_img, aligner, margin=0.8, detect_multiple_faces=True):
    pnet, rnet, onet = aligner
    minsize = 20  # minimum size of face
    threshold = [0.6, 0.7, 0.7]  # three steps's threshold
    factor = 0.709  # scale factor

    if orig_img.ndim < 2:
        return None
    if orig_img.ndim == 2:
        orig_img = to_rgb(orig_img)
    orig_img = orig_img[:, :, 0:3]

    bounding_boxes, _ = detect_face(orig_img, minsize, pnet, rnet, onet, threshold, factor)
    nrof_faces = bounding_boxes.shape[0]
    if nrof_faces > 0:
        det = bounding_boxes[:, 0:4]
        det_arr = []
        img_size = np.asarray(orig_img.shape)[0:2]
        if nrof_faces > 1:
            margin = margin / 1.5
            if detect_multiple_faces:
                for i in range(nrof_faces):
                    det_arr.append(np.squeeze(det[i]))
            else:
                bounding_box_size = (det[:, 2] - det[:, 0]) * (det[:, 3] - det[:, 1])
                img_center = img_size / 2
                offsets = np.vstack([(det[:, 0] + det[:, 2]) / 2 - img_center[1],
                                     (det[:, 1] + det[:, 3]) / 2 - img_center[0]])
                offset_dist_squared = np.sum(np.power(offsets, 2.0), 0)
                index = np.argmax(bounding_box_size - offset_dist_squared * 2.0)  # some extra weight on the centering
                det_arr.append(det[index, :])
        else:
            det_arr.append(np.squeeze(det))
        cropped_arr = []
        bounding_boxes_arr = []
        for i, det in enumerate(det_arr):
            det = np.squeeze(det)
            bb = np.zeros(4, dtype=np.int32)
            side_1 = int((det[2] - det[0]) * margin)
            side_2 = int((det[3] - det[1]) * margin)

            bb[0] = np.maximum(det[0] - side_1 / 2, 0)
            bb[1] = np.maximum(det[1] - side_1 / 2, 0)
            bb[2] = np.minimum(det[2] + side_2 / 2, img_size[1])
            bb[3] = np.minimum(det[3] + side_2 / 2, img_size[0])
            cropped = orig_img[bb[1]:bb[3], bb[0]:bb[2], :]
            cropped_arr.append(cropped)
            bounding_boxes_arr.append([bb[0], bb[1], bb[2], bb[3]])
            # scaled = misc.imresize(cropped, (image_size, image_size), interp='bilinear')
        return cropped_arr, bounding_boxes_arr
    else:
        return None
#
# if __name__ == '__main__':
#     orig_img = misc.imread('orig_img.jpeg')
#     cropped_arr, bounding_boxes_arr = align(orig_img)
#     misc.imsave('test_output.jpeg', cropped_arr[0])
#     print(bounding_boxes_arr)
#



# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function

import argparse
import glob
import os
import random
import sys
import time

import numpy as np

from .differentiator import FawkesMaskGeneration
from .utils import load_extractor, init_gpu, select_target_label, dump_image, reverse_process_cloaked, \
    Faces

random.seed(12243)
np.random.seed(122412)


def generate_cloak_images(sess, feature_extractors, image_X, target_emb=None, th=0.01, faces=None, sd=1e9, lr=2,
                          max_step=500, batch_size=1):
    batch_size = batch_size if len(image_X) > batch_size else len(image_X)

    differentiator = FawkesMaskGeneration(sess, feature_extractors,
                                          batch_size=batch_size,
                                          mimic_img=True,
                                          intensity_range='imagenet',
                                          initial_const=sd,
                                          learning_rate=lr,
                                          max_iterations=max_step,
                                          l_threshold=th,
                                          verbose=1, maximize=False, keep_final=False, image_shape=image_X.shape[1:],
                                          faces=faces)

    cloaked_image_X = differentiator.attack(image_X, target_emb)
    return cloaked_image_X


def check_imgs(imgs):
    if np.max(imgs) <= 1 and np.min(imgs) >= 0:
        imgs = imgs * 255.0
    elif np.max(imgs) <= 255 and np.min(imgs) >= 0:
        pass
    else:
        raise Exception("Image values ")
    return imgs


def main(*argv):
    start_time = time.time()
    if not argv:
        argv = list(sys.argv)

    try:
        import signal
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except Exception as e:
        pass

    parser = argparse.ArgumentParser()
    parser.add_argument('--directory', '-d', type=str,
                        help='directory that contain images for cloaking', default='imgs/')

    parser.add_argument('--gpu', type=str,
                        help='GPU id', default='0')

    parser.add_argument('--mode', type=str,
                        help='cloak generation mode', default='high')
    parser.add_argument('--feature-extractor', type=str,
                        help="name of the feature extractor used for optimization",
                        default="high_extract")

    parser.add_argument('--th', type=float, default=0.01)
    parser.add_argument('--max-step', type=int, default=500)
    parser.add_argument('--sd', type=int, default=1e9)
    parser.add_argument('--lr', type=float, default=2)

    parser.add_argument('--batch-size', type=int, default=1)
    parser.add_argument('--separate_target', action='store_true')

    parser.add_argument('--format', type=str,
                        help="final image format",
                        default="png")
    args = parser.parse_args(argv[1:])

    if args.mode == 'low':
        args.feature_extractor = "high_extract"
        args.th = 0.003
        args.max_step = 100
        args.lr = 15
    elif args.mode == 'mid':
        args.feature_extractor = "high_extract"
        args.th = 0.005
        args.max_step = 100
        args.lr = 15
    elif args.mode == 'high':
        args.feature_extractor = "high_extract"
        args.th = 0.007
        args.max_step = 100
        args.lr = 10
    elif args.mode == 'ultra':
        args.feature_extractor = "high_extract"
        args.th = 0.01
        args.max_step = 1000
        args.lr = 5
    elif args.mode == 'custom':
        pass
    else:
        raise Exception("mode must be one of 'low', 'mid', 'high', 'ultra', 'custom'")

    assert args.format in ['png', 'jpg', 'jpeg']
    if args.format == 'jpg':
        args.format = 'jpeg'

    sess = init_gpu(args.gpu)
    fs_names = [args.feature_extractor]
    feature_extractors_ls = [load_extractor(name) for name in fs_names]

    image_paths = glob.glob(os.path.join(args.directory, "*"))
    image_paths = [path for path in image_paths if "_cloaked" not in path.split("/")[-1]]
    if not image_paths:
        print("No images in the directory")
        exit(1)

    faces = Faces(image_paths, sess, verbose=1)

    orginal_images = faces.cropped_faces
    orginal_images = np.array(orginal_images)

    if args.separate_target:
        target_embedding = []
        for org_img in orginal_images:
            org_img = org_img.reshape([1] + list(org_img.shape))
            tar_emb = select_target_label(org_img, feature_extractors_ls, fs_names)
            target_embedding.append(tar_emb)
        target_embedding = np.concatenate(target_embedding)
    else:
        target_embedding = select_target_label(orginal_images, feature_extractors_ls, fs_names)

    protected_images = generate_cloak_images(sess, feature_extractors_ls, orginal_images,
                                             target_emb=target_embedding, th=args.th, faces=faces, sd=args.sd,
                                             lr=args.lr, max_step=args.max_step, batch_size=args.batch_size)

    faces.cloaked_cropped_faces = protected_images

    cloak_perturbation = reverse_process_cloaked(protected_images) - reverse_process_cloaked(orginal_images)
    final_images = faces.merge_faces(cloak_perturbation)

    for p_img, cloaked_img, path in zip(final_images, protected_images, image_paths):
        file_name = "{}_{}_cloaked.{}".format(".".join(path.split(".")[:-1]), args.mode, args.format)
        dump_image(p_img, file_name, format=args.format)

    elapsed_time = time.time() - start_time
    print('attack cost %f s' % (elapsed_time))


if __name__ == '__main__':
    main(*sys.argv)



# -*- coding: utf-8 -*-
# @Date    : 2020-07-01
# @Author  : Shawn Shan (shansixiong@cs.uchicago.edu)
# @Link    : https://www.shawnshan.com/


__version__ = '0.0.3'

from .differentiator import FawkesMaskGeneration
from .utils import load_extractor, init_gpu, select_target_label, dump_image, reverse_process_cloaked, \
    Faces
from .protection import main
import logging
import sys
import os
logging.getLogger('tensorflow').disabled = True


__all__ = (
    '__version__',
    'FawkesMaskGeneration', 'load_extractor',
    'init_gpu',
    'select_target_label', 'dump_image', 'reverse_process_cloaked', 'Faces', 'main'
)


import os
import re
import sys

from setuptools import setup, Command

__PATH__ = os.path.abspath(os.path.dirname(__file__))

with open("README.md", "r") as fh:
    long_description = fh.read()


def read_version():
    __PATH__ = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(__PATH__, 'fawkes/__init__.py')) as f:
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                                  f.read(), re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find __version__ string")


__version__ = read_version()


# brought from https://github.com/kennethreitz/setup.py
class DeployCommand(Command):
    description = 'Build and deploy the package to PyPI.'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @staticmethod
    def status(s):
        print(s)

    def run(self):

        assert 'dev' not in __version__, (
            "Only non-devel versions are allowed. "
            "__version__ == {}".format(__version__))

        with os.popen("git status --short") as fp:
            git_status = fp.read().strip()
            if git_status:
                print("Error: git repository is not clean.\n")
                os.system("git status --short")
                sys.exit(1)

        try:
            from shutil import rmtree
            self.status('Removing previous builds ...')
            rmtree(os.path.join(__PATH__, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution ...')
        os.system('{0} setup.py sdist'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine ...')
        ret = os.system('twine upload dist/*')
        if ret != 0:
            sys.exit(ret)

        self.status('Creating git tags ...')
        os.system('git tag v{0}'.format(__version__))
        os.system('git tag --list')
        sys.exit()


setup_requires = []

install_requires = [
    'numpy>=1.16.4',
    'tensorflow>=1.13.1',
    'argparse',
    'keras==2.2.5',
    'scikit-image',
    'pillow>=7.0.0',
    'opencv-python>=4.2.0.34',
]

setup(
    name='fawkes',
    version=__version__,
    license='MIT',
    description='An utility to protect user privacy',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/Shawn-Shan/fawkes",
    author='Shawn Shan',
    author_email='shansixiong@cs.uchicago.edu',
    keywords='fawkes privacy clearview',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3',
        'Topic :: System :: Monitoring',
    ],
    packages=['fawkes'],
    install_requires=install_requires,
    setup_requires=setup_requires,
    entry_points={
        'console_scripts': ['fawkes=fawkes:main'],
    },
    cmdclass={
        'deploy': DeployCommand,
    },
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.5',
)
